package com.sleepycat.je.dbi;

import java.io.UnsupportedEncodingException;

import com.sleepycat.je.Environment;
import com.sleepycat.je.DatabaseEntry;
import com.sleepycat.je.DatabaseException;
import com.sleepycat.je.DatabaseUtil;
import com.sleepycat.je.dbi.CursorImpl.SearchMode;
import com.sleepycat.je.tree.MapLN;
import com.sleepycat.je.tree.NameLN;
import com.sleepycat.je.tree.Tree;
import com.sleepycat.je.txn.LockType;
import com.sleepycat.je.txn.Locker;
import com.sleepycat.je.txn.LockerFactory;
import com.sleepycat.je.OperationContext;

/**
 * requires deleteOperation
 */
public privileged aspect TruncateDatabaseOperation {
	/**
	 * To truncate, remove the database named by databaseName and create a new
	 * database in its place.
	 * 
	 * @param returnCount
	 *            if true, must return the count of records in the database,
	 *            which can be an expensive option.
	 */
	long DbTree.truncate(Locker locker, String databaseName, boolean returnCount)
			throws DatabaseException {

		CursorImpl nameCursor = null;
		Locker autoTxn = null;
		try {
			NameLockResult result = lockNameLN(locker, databaseName, "truncate");
			nameCursor = result.nameCursor;
			if (nameCursor == null) {
				return 0;
			} else {

				/*
				 * Make a new database with an empty tree. Make the nameLN refer
				 * to the id of the new database.
				 */
				DatabaseId newId = new DatabaseId(getNextDbId());
				DatabaseImpl newDb = (DatabaseImpl) result.dbImpl.clone();
				newDb.setId(newId);
				newDb.setTree(new Tree(newDb));

				/*
				 * Insert the new MapLN into the id tree. Always use an AutoTxn
				 * on the id databaase, because we can not hold long term locks
				 * on the mapLN.
				 */
				CursorImpl idCursor = null;
				boolean operationOk = false;
				try {
					autoTxn = createLocker(envImpl);
					idCursor = new CursorImpl(idDatabase, autoTxn);
					idCursor.putLN(newId.getBytes(), new MapLN(newDb), false);
					operationOk = true;
				} finally {
					if (idCursor != null) {
						idCursor.close();
					}

					if (autoTxn != null) {
						autoTxn.operationEnd(operationOk);
					}
				}
				result.nameLN.setId(newDb.getId());

				/* If required, count the number of records in the database. */
				long recordCount = 0;
				if (returnCount) {
					recordCount = result.dbImpl.countRecords();
				}

				/* log the nameLN. */
				DatabaseEntry dataDbt = new DatabaseEntry(new byte[0]);
				nameCursor.putCurrent(dataDbt, null, null);

				/*
				 * Marking the lockers should be the last action, since it takes
				 * effect immediately for non-txnal lockers.
				 */

				/* Schedule old database for deletion if txn commits. */
				locker.markDeleteAtTxnEnd(result.dbImpl, true);

				/* Schedule new database for deletion if txn aborts. */
				locker.markDeleteAtTxnEnd(newDb, false);

				return recordCount;
			}

		} catch (CloneNotSupportedException CNSE) {
			throw new DatabaseException(CNSE);
		} finally {
			if (nameCursor != null) {
				nameCursor.close();
			}
		}
	}
	
	
	/**
	 * Truncate a database named by databaseName. Return the new DatabaseImpl
	 * object that represents the truncated database. The old one is marked as
	 * deleted.
	 * 
	 * @deprecated This method used by Database.truncate()
	 */
	TruncateResult DbTree.truncate(Locker locker, DatabaseImpl oldDatabase,
			boolean returnCount) throws DatabaseException {

		CursorImpl nameCursor = new CursorImpl(nameDatabase, locker);

		try {
			String databaseName = getDbName(oldDatabase.getId());
			DatabaseEntry keyDbt = new DatabaseEntry(databaseName
					.getBytes("UTF-8"));
			boolean found = (nameCursor.searchAndPosition(keyDbt, null,
					SearchMode.SET, LockType.WRITE) & CursorImpl.FOUND) != 0;
			if (!found) {

				/*
				 * Should be found, since truncate is instigated from
				 * Database.truncate();
				 */
				throw new DatabaseException("Database " + databaseName
						+ " not found in map tree");
			}

			/* Call getCurrentLN to write lock the nameLN. */
			NameLN nameLN = (NameLN) nameCursor
					.getCurrentLN(LockType.WRITE);
			assert nameLN != null; /* Should be locked. */

			/*
			 * Check the open handle count after we have the write lock and no
			 * other transactions can open. XXX, another handle using the same
			 * txn could open ...
			 */
			int handleCount = oldDatabase.getReferringHandleCount();
			if (handleCount > 1) {
				throw new DatabaseException("Can't truncate database "
						+ databaseName + "," + handleCount
						+ " open databases exist");
			}

			/*
			 * Make a new database with an empty tree. Make the nameLN refer to
			 * the id of the new database.
			 */
			DatabaseImpl newDb;
			DatabaseId newId = new DatabaseId(getNextDbId());
			newDb = (DatabaseImpl) oldDatabase.clone();
			newDb.setId(newId);
			newDb.setTree(new Tree(newDb));

			/* Insert the new db into id -> name map */
			CursorImpl idCursor = null;
			boolean operationOk = false;
			Locker autoTxn = null;
			try {
				autoTxn = createLocker(envImpl);
				idCursor = new CursorImpl(idDatabase, autoTxn);
				idCursor.putLN(newId.getBytes(), new MapLN(newDb), false);
				operationOk = true;
			} finally {
				if (idCursor != null) {
					idCursor.close();
				}

				if (autoTxn != null) {
					autoTxn.operationEnd(operationOk);
				}
			}
			nameLN.setId(newDb.getId());

			/* count records for the deleted database. */
			long count = 0;
			if (returnCount) {
				count = oldDatabase.countRecords();
			}

			/* Schedule database for final deletion during commit. */
			locker.markDeleteAtTxnEnd(oldDatabase, true);

			/* log the nameLN. */
			DatabaseEntry dataDbt = new DatabaseEntry(new byte[0]);
			nameCursor.putCurrent(dataDbt, null, null);

			return new TruncateResult(newDb, (int) count);
		} catch (CloneNotSupportedException CNSE) {
			throw new DatabaseException(CNSE);
		} catch (UnsupportedEncodingException UEE) {
			throw new DatabaseException(UEE);
		} finally {
			nameCursor.close();
		}
	}
	
	
	/**
	 * Truncate a database. Return a new DatabaseImpl object which represents
	 * the new truncated database. The old database is marked as deleted.
	 * 
	 * @deprecated This supports Database.truncate(), which is deprecated.
	 */
	public TruncateResult EnvironmentImpl.truncate(Locker locker, DatabaseImpl database)
			throws DatabaseException {

		return dbMapTree.truncate(locker, database, true);
	}

	/**
	 * Truncate a database.
	 */
	public long EnvironmentImpl.truncate(Locker locker, String databaseName, boolean returnCount)
			throws DatabaseException {

		return dbMapTree.truncate(locker, databaseName, returnCount);
	}
	
	/**
	 * Javadoc for this public method is generated via the doc templates in the
	 * doc_src directory.
	 */
	public long Environment.truncateDatabase(
			OperationContext cxt, String databaseName, boolean returnCount)
			throws DatabaseException {

		checkHandleIsValid();
		checkEnv();
		DatabaseUtil.checkForNullParam(databaseName, "databaseName");

		Locker locker = null;
		boolean operationOk = false;
		long count = 0;
		try {

			/*
			 * Note: use env level isTransactional as proxy for the db
			 * isTransactional.
			 */
			locker = LockerFactory
					.getWritableLocker(this, cxt, true /* retainNonTxnLocks */);

			count = environmentImpl.truncate(locker, databaseName, returnCount);

			operationOk = true;
		} finally {
			if (locker != null) {
				locker.operationEnd(operationOk);
			}
		}
		return count;
	}
}
